@@ -65,7 +65,7 @@ snpgdsPCA <- function(gdsobj, sample.id=NULL, snp.id=NULL,
     nt_old <- blas_get_num_procs()
     if (is.na(nt_old)) nt_old <- 1L
     blas_set_num_threads(ws$num.thread)
-    on.exit(blas_set_num_threads(nt_old))
+    on.exit(blas_set_num_threads(nt_old), add=TRUE)
     # run the C code
     rv <- .Call(gnrPCA, eigen.cnt, algorithm, ws$num.thread, param, verbose)
 
