@@ -1,14 +1,8 @@
 // ===========================================================
-//     _/_/_/   _/_/_/  _/_/_/_/    _/_/_/_/  _/_/_/   _/_/_/
-//      _/    _/       _/             _/    _/    _/   _/   _/
-//     _/    _/       _/_/_/_/       _/    _/    _/   _/_/_/
-//    _/    _/       _/             _/    _/    _/   _/
-// _/_/_/   _/_/_/  _/_/_/_/_/     _/     _/_/_/   _/_/
-// ===========================================================
 //
-// genLD.cpp: Linkage Disequilibrium (LD) analysis on genome-wide association studies
+// genLD.cpp: Linkage Disequilibrium (LD) Analysis on GWAS
 //
-// Copyright (C) 2013	Xiuwen Zheng
+// Copyright (C) 2011-2024    Xiuwen Zheng
 //
 // This file is part of SNPRelate.
 //
@@ -26,11 +20,8 @@
 // If not, see <http://www.gnu.org/licenses/>.
 
 
-// CoreArray library header
-#include <dType.h>
-#include <dVect.h>
-#include <CoreGDSLink.h>
-#include <dGenGWAS.h>
+#ifndef _HEADER_LD_
+#define _HEADER_LD_
 
 // Standard library header
 #include <cmath>
@@ -39,6 +30,10 @@
 #include <list>
 #include <algorithm>
 
+// CoreArray library header
+#include <dGenGWAS.h>
+#include <dVect.h>
+
 
 #ifdef COREARRAY_SIMD_SSE
 #include <xmmintrin.h>
@@ -48,15 +43,11 @@
 #endif
 
 
-#ifndef _FuncLD_H_
-#define _FuncLD_H_
-
 namespace LD
 {
 	// using namespace
 	using namespace std;
 	using namespace CoreArray;
-	using namespace GDSInterface;
 	using namespace GWAS;
 
 
@@ -66,51 +57,52 @@ namespace LD
 	static const int _size = 256*256;
 
 	/// The number of valid pair of SNPs in the packed genotypes
-	UInt8 Valid_Num_SNP[_size];
+	C_UInt8 Valid_Num_SNP[_size];
 	/// The number of aa in a pair of packed SNPs
-	UInt8 Num_aa_SNP[_size];
+	C_UInt8 Num_aa_SNP[_size];
 	/// The number of aA in a pair of packed SNPs
-	UInt8 Num_aA_SNP[_size];
+	C_UInt8 Num_aA_SNP[_size];
 	/// The number of AA in a pair of packed SNPs
-	UInt8 Num_AA_SNP[_size];
+	C_UInt8 Num_AA_SNP[_size];
 
 	/// The number of AA BB pairs in the packed genotypes
-	UInt8 Num_AA_BB_SNP[_size];
+	C_UInt8 Num_AA_BB_SNP[_size];
 	/// The number of AA BB pairs in the packed genotypes
-	UInt8 Num_aa_bb_SNP[_size];
+	C_UInt8 Num_aa_bb_SNP[_size];
 	/// The number of aa BB pairs in the packed genotypes
-	UInt8 Num_aa_BB_SNP[_size];
+	C_UInt8 Num_aa_BB_SNP[_size];
 	/// The number of AA bb pairs in the packed genotypes
-	UInt8 Num_AA_bb_SNP[_size];
+	C_UInt8 Num_AA_bb_SNP[_size];
 	/// The number of double het pairs in the packed genotypes
-	UInt8 Num_DH_SNP[_size];
+	C_UInt8 Num_DH_SNP[_size];
 
 	/// The sum of X for a pair of SNPs in the packed genotypes
-	UInt8 Sum_X_SNP[_size];
+	C_UInt8 Sum_X_SNP[_size];
 	/// The sum of X^2 for a pair of SNPs in the packed genotypes
-	UInt8 Sum_X_2_SNP[_size];
+	C_UInt8 Sum_X_2_SNP[_size];
 	/// The sum of X*Y for a pair of SNPs in the packed genotypes
-	UInt8 Sum_XY_SNP[_size];
+	C_UInt8 Sum_XY_SNP[_size];
 
 	/// The number of haplotype A / A in the packed genotypes
-	UInt8 Num_A_A[_size];
+	C_UInt8 Num_A_A[_size];
 	/// The number of haplotype A / B in the packed genotypes
-	UInt8 Num_A_B[_size];
+	C_UInt8 Num_A_B[_size];
 	/// The number of haplotype B / A in the packed genotypes
-	UInt8 Num_B_A[_size];
+	C_UInt8 Num_B_A[_size];
 	/// The number of haplotype B / B in the packed genotypes
-	UInt8 Num_B_B[_size];
+	C_UInt8 Num_B_B[_size];
 	/// The number of DH (double hets)
-	UInt8 Num_DH2[_size];
+	C_UInt8 Num_DH2[_size];
 
 	/// Genotype, stored in a packed mode
-	auto_ptr<UInt8> PackedGeno;
-	/// the number of samples and snps
-	long nPackedSamp, nSNP;
+	vector<C_UInt8> PackedGeno;
+
+	/// the number of samples and SNPs
+	ssize_t nPackedSamp, NumSNP;
 
 
 	/// initial object
-	class TInit
+	class COREARRAY_DLL_LOCAL TInit
 	{
 	public:
 		TInit()
@@ -154,7 +146,8 @@ namespace LD
 			/// The sum of X*Y for a pair of SNPs in the packed genotypes
 			PACKED_COND((b1 < 3) && (b2 < 3), Sum_XY_SNP, sum += b1*b2);
 
-			static int IncArray[9][5] = {
+			static int IncArray[9][5] =
+			{
 				{ 0, 0, 0, 2, 0 }, // BB, BB
 				{ 0, 0, 1, 1, 0 }, // BB, AB
 				{ 0, 0, 2, 0, 0 }, // BB, AA
@@ -183,18 +176,18 @@ namespace LD
 	/// Composite LD estimation
 	static double PairComposite(const int *snp1, const int *snp2, int cnt)
 	{
-		// Init data
+		// initialize data
 		long n, naa, naA, nAA, nbb, nbB, nBB, nAABB, naabb, naaBB, nAAbb;
 		n = naa = naA = nAA = nbb = nbB = nBB =
 			nAABB = naabb = naaBB = nAAbb = 0;
 
+		// for-loop
 		for (; cnt > 0; cnt--, snp1++, snp2++)
 		{
-			UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
-			UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
+			C_UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
+			C_UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
 			size_t _p = (size_t(g1) << 8) | (g2);
 			size_t _q = (size_t(g2) << 8) | (g1);
-
 			n += Valid_Num_SNP[_p];
 			naa += Num_aa_SNP[_p]; nbb += Num_aa_SNP[_q];
 			naA += Num_aA_SNP[_p]; nbB += Num_aA_SNP[_q];
@@ -217,15 +210,16 @@ namespace LD
 			if (t > 0)
 				return delta / sqrt(t);
 		}
-		return conf_F64_NaN();
+		return R_NaN;
 	}
-	static double PairComposite(const UInt8 *snp1, const UInt8 *snp2)
+	static double PairComposite(const C_UInt8 *snp1, const C_UInt8 *snp2)
 	{
-		// Init data
+		// initialize data
 		long n, naa, naA, nAA, nbb, nbB, nBB, nAABB, naabb, naaBB, nAAbb;
 		n = naa = naA = nAA = nbb = nbB = nBB =
 			nAABB = naabb = naaBB = nAAbb = 0;
 
+		// for-loop
 		for (long cnt=nPackedSamp; cnt > 0; cnt--, snp1++, snp2++)
 		{
 			size_t _p = (size_t(*snp1) << 8) | (*snp2);
@@ -252,7 +246,7 @@ namespace LD
 			if (t > 0)
 				return delta / sqrt(t);
 		}
-		return conf_F64_NaN();
+		return R_NaN;
 	}
 
 	// ---------------------------------------------------------------------
@@ -271,7 +265,8 @@ namespace LD
 		\param pB_B  output the proportion of B / B
 		pA_A + pA_B + pB_A + pB_B = 1.
 	*/
-	inline static void ProportionHaplo(long nA_A, long nA_B, long nB_A, long nB_B, long nDH2,
+	inline static void ProportionHaplo(
+		long nA_A, long nA_B, long nB_A, long nB_B, long nDH2,
 		double &pA_A, double &pA_B, double &pB_A, double &pB_B)
 	{
 		// initial parameters
@@ -346,8 +341,8 @@ namespace LD
 		// Calculate the numbers of haplotypes
 		for (; cnt > 0; cnt--, snp1++, snp2++)
 		{
-			UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
-			UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
+			C_UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
+			C_UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
 			size_t _p = (size_t(g1) << 8) | (g2);
 			nA_A += Num_A_A[_p]; nA_B += Num_A_B[_p];
 			nB_A += Num_B_A[_p]; nB_B += Num_B_B[_p];
@@ -363,13 +358,13 @@ namespace LD
 		double D = pA_A - pA * p_A;
 		return D / sqrt(pA * p_A * pB * p_B);
 	}
-	static double PairR(const UInt8 *snp1, const UInt8 *snp2)
+	static double PairR(const C_UInt8 *snp1, const C_UInt8 *snp2)
 	{
 		// The number of haplotypes, nDH - double hets
 		long nA_A, nA_B, nB_A, nB_B, nDH2;
 		nA_A = nA_B = nB_A = nB_B = nDH2 = 0;
 
-		// Calculate the numbers of haplotypes
+		// for-loop
 		for (long cnt=nPackedSamp; cnt > 0; cnt--, snp1++, snp2++)
 		{
 			size_t _p = (size_t(*snp1) << 8) | (*snp2);
@@ -398,11 +393,11 @@ namespace LD
 		long nA_A, nA_B, nB_A, nB_B, nDH2;
 		nA_A = nA_B = nB_A = nB_B = nDH2 = 0;
 
-		// Calculate the numbers of haplotypes
+		// for-loop
 		for (; cnt > 0; cnt--, snp1++, snp2++)
 		{
-			UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
-			UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
+			C_UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
+			C_UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
 			size_t _p = (size_t(g1) << 8) | (g2);
 			nA_A += Num_A_A[_p]; nA_B += Num_A_B[_p];
 			nB_A += Num_B_A[_p]; nB_B += Num_B_B[_p];
@@ -420,13 +415,13 @@ namespace LD
 		D = D / ((D>=0) ? min(pA*p_B, pB*p_A) : max(-pA*p_A, -pB*p_B));
 		return D;
 	}
-	static double PairDPrime(const UInt8 *snp1, const UInt8 *snp2)
+	static double PairDPrime(const C_UInt8 *snp1, const C_UInt8 *snp2)
 	{
 		// The number of haplotypes, nDH - double hets
 		long nA_A, nA_B, nB_A, nB_B, nDH2;
 		nA_A = nA_B = nB_A = nB_B = nDH2 = 0;
 
-		// Calculate the numbers of haplotypes
+		// for-loop
 		for (long cnt=nPackedSamp; cnt > 0; cnt--, snp1++, snp2++)
 		{
 			size_t _p = (size_t(*snp1) << 8) | (*snp2);
@@ -448,19 +443,20 @@ namespace LD
 		return D;
 	}
 
+
 	// ---------------------------------------------------------------------
-	/// D' by EM algorithm
+	/// Correlation
 	static double PairCorr(const int *snp1, const int *snp2, int cnt)
 	{
-		// Init data
-		long n, X, XX, Y, YY, XY;
+		// initialize data
+		ssize_t n, X, XX, Y, YY, XY;
 		n = X = XX = Y = YY = XY = 0;
 
-		// Calculate the numbers of haplotypes
+		// for-loop
 		for (; cnt > 0; cnt--, snp1++, snp2++)
 		{
-			UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
-			UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
+			C_UInt8 g1 = (0<=*snp1 && *snp1<=2) ? (*snp1 | ~0x03) : 0xFF;
+			C_UInt8 g2 = (0<=*snp2 && *snp2<=2) ? (*snp2 | ~0x03) : 0xFF;
 			size_t _p = (size_t(g1) << 8) | (g2);
 			size_t _q = (size_t(g2) << 8) | (g1);
 			n += Valid_Num_SNP[_p];
@@ -476,19 +472,20 @@ namespace LD
 			if (val > 0)
 				return (XY - double(X)*Y/n) / sqrt(val);
 		}
-		return conf_F64_NaN();
+		return R_NaN;
 	}
-	static double PairCorr(const UInt8 *snp1, const UInt8 *snp2)
+	static double PairCorr(const C_UInt8 *snp1, const C_UInt8 *snp2)
 	{
-		// Init data
-		long n, X, XX, Y, YY, XY;
+		// initialize data
+		ssize_t n, X, XX, Y, YY, XY;
 		n = X = XX = Y = YY = XY = 0;
 
-		// Calculate the numbers of haplotypes
-		for (long cnt=nPackedSamp; cnt > 0; cnt--, snp1++, snp2++)
+		// for-loop
+		for (size_t cnt=nPackedSamp; cnt > 0; cnt--)
 		{
 			size_t _p = (size_t(*snp1) << 8) | (*snp2);
 			size_t _q = (size_t(*snp2) << 8) | (*snp1);
+			snp1++; snp2++;
 			n += Valid_Num_SNP[_p];
 			X += Sum_X_SNP[_p]; Y += Sum_X_SNP[_q];
 			XX += Sum_X_2_SNP[_p]; YY += Sum_X_2_SNP[_q];
@@ -502,16 +499,37 @@ namespace LD
 			if (val > 0)
 				return (XY - double(X)*Y/n) / sqrt(val);
 		}
-		return conf_F64_NaN();
+		return R_NaN;
+	}
+
+
+
+	// ---------------------------------------------------------------------
+	/// Covariance
+	static double PairCov(const C_UInt8 *snp1, const C_UInt8 *snp2)
+	{
+		// initialize data
+		ssize_t n=0, X=0, Y=0, XY=0;
+
+		// for-loop
+		for (size_t cnt=nPackedSamp; cnt > 0; cnt--)
+		{
+			size_t _p = (size_t(*snp1) << 8) | (*snp2);
+			size_t _q = (size_t(*snp2) << 8) | (*snp1);
+			snp1 ++; snp2 ++;
+			n += Valid_Num_SNP[_p];
+			X += Sum_X_SNP[_p]; Y += Sum_X_SNP[_q];
+			XY += Sum_XY_SNP[_p];
+		}
+
+		return (n > 1) ? (XY - double(X)*Y/n) / (n - 1) : R_NaN;
 	}
 
 
 
 
-
-
 	// ---------------------------------------------------------------------
-	// public parameters
+	// Public parameters
 
 	// the method of Linkage Disequilibrium (LD)
 	// 1 -- "composite"  Composite LD coefficients (by default)
@@ -520,33 +538,34 @@ namespace LD
 	// 4 -- "corr"       Correlation coefficient (BB, AB, AA are codes as 0, 1, 2)
 	int LD_Method = 1;
 
-	/// initialize the variable "Geno" with packed genotypes
-	void InitPackedGeno()
+	/// Initialize the variable "Geno" with packed genotypes
+	COREARRAY_DLL_LOCAL void InitPackedGeno()
 	{
 		// set # of samples and snps
-		nSNP = MCWorkingGeno.Space.SNPNum();
-		nPackedSamp = (MCWorkingGeno.Space.SampleNum() % 4 > 0) ?
-			(MCWorkingGeno.Space.SampleNum()/4 + 1) : (MCWorkingGeno.Space.SampleNum()/4);
-		PackedGeno.reset(new UInt8[nPackedSamp*nSNP]);
+		NumSNP = MCWorkingGeno.Space().SNPNum();
+		nPackedSamp = (MCWorkingGeno.Space().SampleNum() % 4 > 0) ?
+			(MCWorkingGeno.Space().SampleNum()/4 + 1) :
+			(MCWorkingGeno.Space().SampleNum()/4);
+		PackedGeno.resize(nPackedSamp * NumSNP);
 
 		// buffer
-		CdBufSpace Buf(MCWorkingGeno.Space, true, CdBufSpace::acInc);
-		UInt8 *p = PackedGeno.get();
-		for (long i=0; i < MCWorkingGeno.Space.SNPNum(); i++)
+		CdBufSpace Buf(MCWorkingGeno.Space(), true, CdBufSpace::acInc);
+		C_UInt8 *p = &PackedGeno[0];
+		for (long i=0; i < MCWorkingGeno.Space().SNPNum(); i++)
 		{
 			p = Buf.ReadPackedGeno(i, p);
 		}
 	}
 
-	/// destroy the variable "Geno" with packed genotypes
-	void DonePackedGeno()
+	/// Destroy the variable "Geno" with packed genotypes
+	COREARRAY_DLL_LOCAL void DonePackedGeno()
 	{
-		PackedGeno.reset(NULL);
+		PackedGeno.clear();
 	}
 
 
-	// to compute pair LD
-	double calcLD(const int *snp1, const int *snp2, int cnt,
+	/// Compute pair LD
+	COREARRAY_DLL_LOCAL double Compute_LD(const int *snp1, const int *snp2, int cnt,
 		double &pA_A, double &pA_B, double &pB_A, double &pB_B)
 	{
 		switch (LD_Method)
@@ -560,155 +579,302 @@ namespace LD
 			case 4:
 				return PairCorr(snp1, snp2, cnt);
 		}
-		return conf_F64_NaN();
+		return R_NaN;
 	}
 
-	// to compute LD matrix (n x n)
-	void calcLD_mat(int nThread, double *out_LD)
+
+
+	/// The thread entry for the calculation of LD measure
+	static void Entry_LD_Matrix(PdThread Thread, int ThreadIndex, void *Param)
 	{
-		for (int i=0; i < nSNP; i++)
+		double *base = (double*)Param;
+		IdMatTri I = Array_Thread_MatIdx[ThreadIndex];
+		C_UInt8 *pGeno = &PackedGeno[0];
+
+		for (C_Int64 n=Array_Thread_MatCnt[ThreadIndex]; n > 0; n--)
 		{
-			out_LD[i*nSNP+i] = 1;
-			for (int j=i+1; j < nSNP; j++)
+			const int i = I.Row(), j = I.Column();
+			double &p1 = base[i*NumSNP + j];
+			double &p2 = base[j*NumSNP + i];
+			C_UInt8 *s1 = pGeno + i*nPackedSamp;
+			C_UInt8 *s2 = pGeno + j*nPackedSamp;
+
+			switch (LD_Method)
 			{
-				double &p1 = out_LD[i*nSNP+j], &p2 = out_LD[j*nSNP+i];
-				UInt8 *s1 = PackedGeno.get() + i*nPackedSamp;
-				UInt8 *s2 = PackedGeno.get() + j*nPackedSamp;
-				switch (LD_Method)
+				case 1:
+					p1 = p2 = PairComposite(s1, s2); break;
+				case 2:
+					p1 = p2 = PairR(s1, s2); break;
+				case 3:
+					p1 = p2 = PairDPrime(s1, s2); break;
+				case 4:
+					p1 = p2 = PairCorr(s1, s2); break;
+				case 5:
+					p1 = p2 = PairCov(s1, s2); break;
+				default:
+					p1 = p2 = R_NaN;
+			}
+
+			++ I;
+		}
+	}
+
+	/// Calculate LD matrix (n x n)
+	COREARRAY_DLL_LOCAL void Compute_LD_Matrix(int nThread, double *out_LD)
+	{
+		Array_SplitJobs(nThread, NumSNP,
+			Array_Thread_MatIdx, Array_Thread_MatCnt);
+
+		GDS_Parallel_RunThreads(Entry_LD_Matrix, out_LD, nThread);
+	}
+
+
+
+	struct TParam_LD_SlideMat
+	{
+		int Num_Thread;
+		int Num_Slide;
+		bool TrimMat;
+		C_Int64 TotalCount;
+		double *OutMatrix;
+	};
+
+	/// The thread entry for the calculation of LD measure
+	static void Entry_LD_SlideMat(PdThread Thread, int ThreadIndex, void *Param)
+	{
+		TParam_LD_SlideMat *pm = (TParam_LD_SlideMat*)Param;
+		vector<C_Int64> StartIndex(pm->Num_Thread);
+		Array_SplitJobs(pm->Num_Thread, pm->TotalCount, &StartIndex[0],
+			Array_Thread_MatCnt);
+
+		// starting point
+		C_Int64 st = StartIndex[ThreadIndex];
+		ssize_t snp_i, snp_inc;
+		double *pVal;
+		if (pm->TrimMat)
+		{
+			snp_i = st / pm->Num_Slide;
+			snp_inc = st % pm->Num_Slide + 1;
+			pVal = pm->OutMatrix + (snp_i * pm->Num_Slide) + (snp_inc-1);
+		} else {
+			snp_i = 0; snp_inc = 1;
+			pVal = pm->OutMatrix;
+			for (; st > 0; st--)
+			{
+				pVal ++; snp_inc ++;
+				if ((snp_inc > pm->Num_Slide) || (snp_i+snp_inc >= NumSNP))
 				{
-					case 1:
-						p1 = p2 = PairComposite(s1, s2); break;
-					case 2:
-						p1 = p2 = PairR(s1, s2); break;
-					case 3:
-						p1 = p2 = PairDPrime(s1, s2); break;
-					case 4:
-						p1 = p2 = PairCorr(s1, s2); break;
-					default:
-						p1 = p2 = conf_F64_NaN();
+					snp_inc = 1;
+					snp_i ++;
+					pVal = pm->OutMatrix + (snp_i * pm->Num_Slide) + (snp_inc-1);
 				}
 			}
 		}
+
+		C_UInt8 *pGeno = &PackedGeno[0];
+		for (C_Int64 n=Array_Thread_MatCnt[ThreadIndex]; n > 0; n--)
+		{
+			C_UInt8 *s1 = pGeno + snp_i * nPackedSamp;
+			C_UInt8 *s2 = pGeno + (snp_i+snp_inc) * nPackedSamp;
+			switch (LD_Method)
+			{
+				case 1:
+					*pVal = PairComposite(s1, s2); break;
+				case 2:
+					*pVal = PairR(s1, s2); break;
+				case 3:
+					*pVal = PairDPrime(s1, s2); break;
+				case 4:
+					*pVal = PairCorr(s1, s2); break;
+				case 5:
+					*pVal = PairCov(s1, s2); break;
+				default:
+					*pVal = R_NaN;
+			}
+
+			pVal ++; snp_inc ++;
+			if ((snp_inc > pm->Num_Slide) || (snp_i+snp_inc >= NumSNP))
+			{
+				snp_inc = 1;
+				snp_i ++;
+				pVal = pm->OutMatrix + (snp_i * pm->Num_Slide) + (snp_inc-1);
+			}
+		}
 	}
 
-	// to compute LD matrix (n_slide x n)
-	void calcLD_slide_mat(int nThread, double *out_LD, int n_slide)
+	/// Compute LD matrix (n_slide x n)
+	COREARRAY_DLL_LOCAL void Compute_LD_Slide_Matrix(int nThread,
+		double *out_LD, int n_slide, bool trim_mat)
 	{
-		for (int i=0; i < nSNP; i++)
-		{
-			for (int j=i+1; (j < nSNP) && ((j-i) <= n_slide); j++)
-			{
-				int d = j - i - 1;
-				double &pVal = out_LD[i*n_slide + d];				
-				UInt8 *s1 = PackedGeno.get() + i*nPackedSamp;
-				UInt8 *s2 = PackedGeno.get() + j*nPackedSamp;
-				switch (LD_Method)
-				{
-					case 1:
-						pVal = PairComposite(s1, s2); break;
-					case 2:
-						pVal = PairR(s1, s2); break;
-					case 3:
-						pVal = PairDPrime(s1, s2); break;
-					case 4:
-						pVal = PairCorr(s1, s2); break;
-					default:
-						pVal = conf_F64_NaN();
-				}
-			}
-		}
+		TParam_LD_SlideMat param;
+		param.Num_Thread = nThread;
+		param.Num_Slide = n_slide;
+		param.TrimMat = trim_mat;
+		param.TotalCount = trim_mat ? (C_Int64)(NumSNP - n_slide) * n_slide :
+			(C_Int64)NumSNP * n_slide - n_slide*(n_slide+1)/2;
+		param.OutMatrix = out_LD;
+
+		GDS_Parallel_RunThreads(Entry_LD_SlideMat, &param, nThread);
 	}
 
 
 
-	// ---------------------------------------------------------------------
-	// to prune SNPs
+	// =====================================================================
+	// Prune SNPs
+	// =====================================================================
 
 	struct TSNP
 	{
 		int idx, pos_bp;
-		vector<UInt8> genobuf;
-		TSNP(int n=0): genobuf(n) {}
-		TSNP(int _idx, int _pos, int n=0): genobuf(n) { idx = _idx; pos_bp = _pos; }
+		vector<C_UInt8> genobuf;
+		TSNP(int n=0): genobuf(n)
+			{ idx = pos_bp = 0; }
+		TSNP(int _idx, int _pos, int n=0): genobuf(n)
+			{ idx = _idx; pos_bp = _pos; }
 	};
 
-	// to compute pair LD
-	static double _CalcLD(const UInt8 *snp1, const UInt8 *snp2)
+	// Compute pair LD
+	static double _CalcLD(const C_UInt8 *snp1, const C_UInt8 *snp2)
 	{
 		switch (LD_Method)
 		{
-			case 1:
-				return PairComposite(snp1, snp2);
-			case 2:
-				return PairR(snp1, snp2);
-			case 3:
-				return PairDPrime(snp1, snp2);
-			case 4:
-				return PairCorr(snp1, snp2);
+			case 1:  return PairComposite(snp1, snp2);
+			case 2:  return PairR(snp1, snp2);
+			case 3:  return PairDPrime(snp1, snp2);
+			case 4:  return PairCorr(snp1, snp2);
+			default: return R_NaN;
 		}
-		return conf_F64_NaN();
 	}
 
-	void calcLDpruning(int StartIdx, int *pos_bp, int slide_max_bp, int slide_max_n,
-		const double LD_threshold, bool *out_SNP)
+	class COREARRAY_DLL_DEFAULT CThreadPoolLD
+	{
+	public:
+		const C_UInt8 *BaseGeno;
+		vector<const C_UInt8*> List;
+		CThreadPoolEx<CThreadPoolLD> thpool;
+
+		/// constructor
+		CThreadPoolLD(int num_thread, const C_UInt8 *baseG, double ld_cutoff):
+			BaseGeno(baseG), thpool(num_thread), nthread(num_thread),
+			thread_start(num_thread), thread_length(num_thread),
+			test_flag(num_thread), LD_threshold(ld_cutoff)
+		{ }
+
+		inline void TestLD(const C_UInt8 geno[], bool &to_include)
+		{
+			if (to_include)
+			{
+				if (nthread <= 1)
+				{
+					if (fabs(_CalcLD(geno, BaseGeno)) > LD_threshold)
+						to_include = false;
+				} else
+					List.push_back(geno);
+			}
+		}
+
+		inline void TestLD_finish(bool &to_include)
+		{
+			if (to_include && nthread>1 && !List.empty())
+			{
+				thpool.Split(nthread, List.size(), &thread_start[0], &thread_length[0]);
+				thpool.BatchWork(this, &CThreadPoolLD::thread_calc_LD, nthread);
+				for (int i=0; i < nthread; i++)
+				{
+					if (test_flag[i]) { to_include = false; break; }
+				}
+			}
+		}
+
+	private:
+		int nthread;
+		vector<size_t> thread_start, thread_length;
+		vector<int> test_flag;
+		double LD_threshold;
+
+		void thread_calc_LD(size_t i, size_t num)
+		{
+			test_flag[i] = false;
+			size_t st = thread_start[i];
+			for (size_t n = thread_length[i]; n > 0; n--)
+			{
+				if (fabs(_CalcLD(List[st++], BaseGeno)) > LD_threshold)
+				{
+					test_flag[i] = true; break;
+				}
+			}
+
+		}
+	};
+
+	COREARRAY_DLL_LOCAL void Perform_LD_Pruning(int StartIdx, int *pos_bp,
+		int slide_max_bp, int slide_max_n, const double LD_threshold,
+		C_BOOL *out_SNP, int num_thread, bool verbose)
 	{
 		// initial variables
-		nPackedSamp = (MCWorkingGeno.Space.SampleNum() % 4 > 0) ?
-			(MCWorkingGeno.Space.SampleNum()/4 + 1) : (MCWorkingGeno.Space.SampleNum()/4);
+		nPackedSamp = (MCWorkingGeno.Space().SampleNum() % 4 > 0) ?
+			(MCWorkingGeno.Space().SampleNum()/4 + 1) :
+			(MCWorkingGeno.Space().SampleNum()/4);
 		list<TSNP> ListGeno;
 		out_SNP[StartIdx] = true;
-		auto_ptr<UInt8> buf(new UInt8[nPackedSamp]);
+		vector<C_UInt8> buf(nPackedSamp);
 
 		// -----------------------------------------------------
-		// increasing searching
+		// increasing searching: i --> i+1
 
-		CdBufSpace BufSNP(MCWorkingGeno.Space, true, CdBufSpace::acInc);
+		CdBufSpace BufSNP(MCWorkingGeno.Space(), true, CdBufSpace::acInc);
 		ListGeno.push_back(TSNP(StartIdx, pos_bp[StartIdx], nPackedSamp));
 		BufSNP.ReadPackedGeno(StartIdx, &(ListGeno.back().genobuf[0]));
+		CdProgression progress(2, verbose);
+		if (verbose) Rprintf("|");
+
+		CThreadPoolLD SetLD(num_thread, &buf[0], LD_threshold);
 
 		// for-loop, increasing
-		for (int i=StartIdx+1; i < MCWorkingGeno.Space.SNPNum(); i++)
+		progress.Init(MCWorkingGeno.Space().SNPNum()-StartIdx-1, verbose);
+		for (int i=StartIdx+1; i < MCWorkingGeno.Space().SNPNum(); i++)
 		{
 			// load genotypes
-			BufSNP.ReadPackedGeno(i, buf.get());
+			BufSNP.ReadPackedGeno(i, &buf[0]);
+			progress.Forward(1, verbose);
 			// detect LD
-			int TotalCnt = 0, ValidCnt = 0;
-			list<TSNP>::iterator it;
-			for (it=ListGeno.begin(); it != ListGeno.end(); TotalCnt++)
+			SetLD.List.clear();
+			bool to_include = true;
+			for (list<TSNP>::iterator it=ListGeno.begin(); it != ListGeno.end(); )
 			{
 				// check whether it is in the sliding window
 				if ((abs(i - it->idx) <= slide_max_n) &&
 					(abs(pos_bp[i] - it->pos_bp) <= slide_max_bp))
 				{
-					if (fabs(_CalcLD(&(it->genobuf[0]), buf.get())) <= LD_threshold)
-						ValidCnt ++;
-					it++;
+					SetLD.TestLD(&(it->genobuf[0]), to_include);
+					it ++;
 				} else {
-					ValidCnt ++;
 					// delete it
 					list<TSNP>::iterator tmp_it = it;
-					it++;
+					it ++;
 					ListGeno.erase(tmp_it);
 				}
 			}
+			SetLD.TestLD_finish(to_include);
 			// handle
-			out_SNP[i] = (ValidCnt == TotalCnt);
+			out_SNP[i] = to_include;
 			if (out_SNP[i])
 			{
 				ListGeno.push_back(TSNP(i, pos_bp[i], nPackedSamp));
-				memmove(&(ListGeno.back().genobuf[0]), buf.get(), nPackedSamp);
+				memmove(&(ListGeno.back().genobuf[0]), &buf[0], nPackedSamp);
 			}
 		}
 
 		// -----------------------------------------------------
-		// decreasing searching
+		// decreasing searching: i --> i-1
 
+		if (verbose) Rprintf("|");
 		ListGeno.clear();
-		for (int i=StartIdx; i < MCWorkingGeno.Space.SNPNum(); i++)
+		for (int i=StartIdx; i < MCWorkingGeno.Space().SNPNum(); i++)
 		{
 			if (out_SNP[i])
 			{
-				// check whether it is in the sliding window
+				// check whether it is in the sliding window of StartIdx
 				if ((abs(i - StartIdx) <= slide_max_n) &&
 					(abs(pos_bp[i] - pos_bp[StartIdx]) <= slide_max_bp))
 				{
@@ -720,41 +886,154 @@ namespace LD
 		}
 
 		BufSNP.SetAccessFlag(CdBufSpace::acDec);
-		// for-loop, descreasing
+		progress.Init(StartIdx-1, verbose);
+		// for-loop, decreasing
 		for (int i=StartIdx-1; i >= 0; i--)
 		{
 			// load genotypes
-			BufSNP.ReadPackedGeno(i, buf.get());
+			BufSNP.ReadPackedGeno(i, &buf[0]);
+			progress.Forward(1, verbose);
 			// detect LD
-			int TotalCnt = 0, ValidCnt = 0;
-			list<TSNP>::iterator it;
-			for (it=ListGeno.begin(); it != ListGeno.end(); TotalCnt++)
+			SetLD.List.clear();
+			bool to_include = true;
+			for (list<TSNP>::iterator it=ListGeno.begin(); it != ListGeno.end(); )
 			{
 				// check whether it is in the sliding window
 				if ((abs(i - it->idx) <= slide_max_n) &&
 					(abs(pos_bp[i] - it->pos_bp) <= slide_max_bp))
 				{
-					if (fabs(_CalcLD(&(it->genobuf[0]), buf.get())) <= LD_threshold)
-						ValidCnt ++;
+					SetLD.TestLD(&(it->genobuf[0]), to_include);
 					it++;
 				} else {
-					ValidCnt ++;
 					// delete it
 					list<TSNP>::iterator tmp_it = it;
 					it++;
 					ListGeno.erase(tmp_it);
 				}
 			}
+			SetLD.TestLD_finish(to_include);
 			// handle
-			out_SNP[i] = (ValidCnt == TotalCnt);
+			out_SNP[i] = to_include;
 			if (out_SNP[i])
 			{
 				ListGeno.push_front(TSNP(i, pos_bp[i], nPackedSamp));
-				memmove(&(ListGeno.front().genobuf[0]), buf.get(), nPackedSamp);
+				memmove(&(ListGeno.front().genobuf[0]), &buf[0], nPackedSamp);
 			}
 		}
+
+		if (verbose) Rprintf("|\n");
 	}
 }
 
 
-#endif  /* _FuncLD_H_ */
+using namespace LD;
+
+extern "C"
+{
+// =======================================================================
+// the functions for linkage disequilibrium (LD)
+//
+
+/// the functions for Linkage Disequilibrium (LD) analysis
+COREARRAY_DLL_EXPORT SEXP gnrLDpair(SEXP snp1, SEXP snp2, SEXP method)
+{
+	COREARRAY_TRY
+
+		double pA_A, pA_B, pB_A, pB_B;
+		LD::LD_Method = Rf_asInteger(method);
+
+		PROTECT(rv_ans = NEW_NUMERIC(5));
+		double *p = REAL(rv_ans);
+		p[0] = LD::Compute_LD(INTEGER(snp1), INTEGER(snp2),
+			XLENGTH(snp1), pA_A, pA_B, pB_A, pB_B);
+		p[1] = pA_A; p[2] = pA_B; p[3] = pB_A; p[4] = pB_B;
+		UNPROTECT(1);
+
+	COREARRAY_CATCH
+}
+
+
+/// Calculate LD coefficients
+COREARRAY_DLL_EXPORT SEXP gnrLDMat(SEXP method, SEXP NumSlide, SEXP MatTrim,
+	SEXP NumThread, SEXP Verbose)
+{
+	int n_slide = Rf_asInteger(NumSlide);
+	int trim_flag = Rf_asLogical(MatTrim);
+	if (trim_flag == NA_INTEGER)
+		Rf_error("'mat.trim' should be TRUE or FALSE");
+	int nThread = Rf_asInteger(NumThread);
+	if (nThread <= 0)
+		Rf_error("Invalid 'num.thread'.");
+	bool verbose = SEXP_Verbose(Verbose);
+
+	COREARRAY_TRY
+
+		// ======== To cache the genotype data ========
+		CachingSNPData("LD matrix", verbose);
+
+		// initialize the packed genotypes
+		LD::InitPackedGeno();
+		LD::LD_Method = Rf_asInteger(method);
+		R_xlen_t nSNP = MCWorkingGeno.Space().SNPNum();
+
+		if (n_slide <= 0)
+		{
+			// n-by-n matrix
+			PROTECT(rv_ans = Rf_allocMatrix(REALSXP, nSNP, nSNP));
+
+			LD::Compute_LD_Matrix(nThread, REAL(rv_ans));
+
+		} else {
+			// sliding windows
+			if (n_slide > nSNP) n_slide = nSNP;
+			if (trim_flag)
+			{
+				PROTECT(rv_ans = Rf_allocMatrix(REALSXP, n_slide,
+					MCWorkingGeno.Space().SNPNum() - n_slide));
+			} else {
+				PROTECT(rv_ans = Rf_allocMatrix(REALSXP, n_slide,
+					MCWorkingGeno.Space().SNPNum()));
+				{
+					double *p = REAL(rv_ans);
+					for (R_xlen_t n=XLENGTH(rv_ans); n > 0; n--)
+						*p ++ = R_NaN;
+				}
+			}
+
+			LD::Compute_LD_Slide_Matrix(nThread, REAL(rv_ans), n_slide,
+				trim_flag);
+		}
+
+		UNPROTECT(1);
+
+	COREARRAY_CATCH
+}
+
+
+/// Prune SNPs based on LD
+COREARRAY_DLL_EXPORT SEXP gnrLDpruning(SEXP StartIdx, SEXP pos_bp,
+	SEXP slide_max_bp, SEXP slide_max_n, SEXP LD_threshold, SEXP method,
+	SEXP NumThread, SEXP verbose)
+{
+	COREARRAY_TRY
+
+		vector<C_BOOL> flag(MCWorkingGeno.Space().SNPNum());
+		const int nthread = Rf_asInteger(NumThread);
+		LD::LD_Method = Rf_asInteger(method);
+		LD::Perform_LD_Pruning(Rf_asInteger(StartIdx)-1, INTEGER(pos_bp),
+			Rf_asInteger(slide_max_bp), Rf_asInteger(slide_max_n),
+			Rf_asReal(LD_threshold), &flag[0],
+			nthread, Rf_asLogical(verbose) == TRUE);
+
+		PROTECT(rv_ans = NEW_LOGICAL(MCWorkingGeno.Space().SNPNum()));
+		int *p = LOGICAL(rv_ans);
+		for (long i=0; i < MCWorkingGeno.Space().SNPNum(); i++)
+			p[i] = flag[i] ? TRUE : FALSE;
+		UNPROTECT(1);
+
+	COREARRAY_CATCH
+}
+
+}
+
+#endif  /* _HEADER_LD_ */
